from typing import List
import os
import logging
from xml.etree import ElementTree as ET
from xmlschema import XMLSchema
from utils.database import DatabaseManager
from utils.errors.import_error.xsd_scheme_not_found import XsdSchemeNotFoundError
from utils.errors.import_error.invalid_contents_error import InvalidContentsError
from utils.errors.import_error.import_error import ImportError
from models.book import Book
from models.author import Author
from models.book_category import BookCategory
from sqlalchemy.exc import IntegrityError

class BookImporter:
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        try:
            self.logger.debug("Opening XSD scheme in ./")
            scheme_path = os.path.join(os.path.dirname(__file__), "book_import_scheme.xsd")
            self.schema = XMLSchema(scheme_path)
        except Exception as e:
            self.logger.error("Failed to load XSD scheme")
            raise XsdSchemeNotFoundError(f"Failed to load XSD schema: {e}")

    def parse_xml(self, file_path: str) -> List[Book]:
        """Parses the XML file and validates it against the XSD schema."""
        try:
            tree = ET.parse(file_path)
            root = tree.getroot()

            if not self.schema.is_valid(file_path):
                raise InvalidContentsError("XML file is not valid according to XSD schema.")

            books = []
            for book_element in root.findall("book"):
                title = book_element.find("title").text
                year_published = book_element.find("year_published").text
                description = book_element.find("description").text
                isbn = book_element.find("isbn").text

                # Parse author
                author_element = book_element.find("author")
                first_name = author_element.find("first_name").text
                last_name = author_element.find("last_name").text
                author = Author(first_name=first_name, last_name=last_name)

                # Parse categories
                category_elements = book_element.find("categories").findall("category")
                categories = [BookCategory(name=category_element.text) for category_element in category_elements]

                # Create a Book object
                book = Book(
                    title=title,
                    description=description,
                    year_published=year_published,
                    isbn=isbn,
                    author=author,
                    categories=categories
                )
                books.append(book)

            return books
        except ET.ParseError as e:
            raise ImportError(f"Failed to parse XML file: {e}")

    def filter_new_books(self, books: List[Book]) -> List[Book]:
        """Filters out books that already exist in the database."""
        new_books = []
        with DatabaseManager.get_session() as session:
            for book in books:
                existing_book = session.query(Book).filter(
                    Book.isbn == book.isbn,
                ).first()
                if existing_book is None:
                    new_books.append(book)
        return new_books

    def save_books(self, books: List[Book]):
        """Saves a list of books to the database."""
        try:
            with DatabaseManager.get_session() as session:
                for book in books:
                    # Check if the author exists, otherwise add
                    existing_author = session.query(Author).filter_by(
                        first_name=book.author.first_name,
                        last_name=book.author.last_name
                    ).first()
                    if existing_author:
                        book.author = existing_author
                    else:
                        session.add(book.author)
                        session.commit()

                    # Handle categories
                    new_categories = []
                    for category in book.categories:
                        existing_category = session.query(BookCategory).filter_by(name=category.name).first()
                        if existing_category:
                            new_categories.append(existing_category)
                        else:
                            self.logger.debug(f"Adding new category: {category.name}")
                            session.add(category)
                            session.commit()
                            new_categories.append(category)
                    # Replace book categories with the resolved categories
                    book.categories = new_categories

                    # Check if the book already exists
                    existing_book = session.query(Book).filter_by(isbn=book.isbn).first()
                    if not existing_book:
                        session.add(book)
                    else:
                        self.logger.warning(f"Book with ISBN {book.isbn} already exists. Skipping.")

                # Commit all changes
                session.commit()
        except IntegrityError as e:
            raise ImportError(f"An error occurred when importing books: {e}") from e
